#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

int main (int, char *[]);
void readFids (void);
void readFidsDir (char *, int);
void readInfo (int, unsigned long);
void strupper (char *);
void mirrorPlaylist (char *, unsigned long);
void sanatize (unsigned char *);

// This function should be in stdio.h, but isn't
ssize_t getline (char **linepte, size_t *n, FILE *stream);

#define TYPE_NON_EXISTANT (-1)
#define TYPE_PLAYLIST      (0)
#define TYPE_TUNE          (1)

#define DISK_NON_EXISTANT (-1)
#define DISK_DRIVE_0       (0)
#define DISK_DRIVE_1       (1)

#define TYPE_AUDIO (0)
#define TYPE_INFO  (1)

#define MAX_FIDS (65536)

#define VERSION "0.2.2"

struct FID {
	int  type;
	int  disk;
	int  track;
	long duration;
	char *title;
	char *artist;
	char *source;
	char *codec;
};

struct FID fids[MAX_FIDS];

int main (int argc, char *argv[])
{
	char *dir;
	struct stat filestat;
	
	printf ("mirrordb v%s\n", VERSION);
	
	if (argc < 2) {
		printf ("No directory specified, using /drive0/mirror\n");
		dir = "/drive0/mirror";
	} else if (argc == 2)
		dir = argv[1];
	else {
		fprintf (stderr, "Wrong number of arguments\n");
		return 1;
	}
	
	errno = 0;
	stat (dir, &filestat);
	if (errno == ENOENT) {
		printf ("Directory %s does not exist, creating it\n",
			dir);
		if (mkdir (dir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH |
			S_IXOTH) == -1) {
			perror ("Error creating directory");
			return 1;
		}
	}
	readFids ();
	printf ("Mirroring playlists\n");
	mirrorPlaylist (dir, 16);
	printf ("Playlists mirrored\n");
	
	return 0;
}

void readFids (void)
{
	unsigned long i;
	for (i = 0; i < MAX_FIDS; i++) {
		fids[i].type = TYPE_NON_EXISTANT;
		fids[i].disk = DISK_NON_EXISTANT;
		fids[i].track = -1;
		fids[i].duration = 0;
		fids[i].title = NULL;
		fids[i].artist = NULL;
		fids[i].source = NULL;
		fids[i].codec = NULL;
	}
	
	printf ("Reading fids\n");
	readFidsDir ("/drive0/fids", DISK_DRIVE_0);
	readFidsDir ("/drive1/fids", DISK_DRIVE_1);
	printf ("Fids read\n");
}

void readFidsDir (char *dir, int disk)
{
	DIR *fidsDir;
	struct dirent *entry;
	int type;
	
	fidsDir = opendir (dir);
	if (fidsDir == NULL) {
		if (disk != 0)
			return;
		else {
			perror ("Error while opening fids directory");
			exit (1);
		}
	}
	errno = 0;
	while ((entry = readdir (fidsDir)) != NULL) {
		unsigned long fid;
		
		// Try to convert name to integer, fail silently if it
		// isn't one.
		errno = 0;
		fid = strtol (entry->d_name, NULL, 16);
		if (errno || (! fid))
			continue;
		
		if (fid & 0x1)
			type = TYPE_INFO;
		else
			type = TYPE_AUDIO;
		fid = fid >> 4;
		if (fid > MAX_FIDS) {
			fprintf (stderr, "Error: FID too large\n");
			exit (1);
		}
		
		if (type == TYPE_INFO)
			readInfo (disk, fid);
		else
			fids[fid].disk = disk;

		errno = 0;
	}
	if (errno) {
		perror ("Error while reading fids directory");
		exit (1);
	}
	closedir (fidsDir);
}

void readInfo (int disk, unsigned long fid)
{
	char filename[32];
	FILE *file;
	char *line;
	size_t n;
	
	sprintf (filename, "/drive%d/fids/%lx1", disk, fid);
	file = fopen (filename, "r");
	line = NULL;
	n = 0;
	while (getline (&line, &n, file) > 0)
	{
		char *label, *p, *value, *c;
		
		// Split line along first '=' into label and value
		label = line;
		p = strchr (line, '=');
		if (p == NULL)
		{
			fprintf (stderr, "Warning: skipping invalid "
				"line '%s' in /drive%d/fids/%lx1\n",
				line, disk, fid);
			continue;
		}
		*p = '\0';
		value = p + 1;

		// Remove linefeed, if any, from end of value
		c = value + strlen (value) - 1;
		if (*c == '\n')
			*c = '\0';

		strupper (label);
		if (strcmp (label, "TITLE") == 0)
			fids[fid].title = strdup (value);
		else if (strcmp (label, "ARTIST") == 0)
			fids[fid].artist = strdup (value);
		else if (strcmp (label, "SOURCE") == 0)
			fids[fid].source = strdup (value);
		else if (strcmp (label, "TRACKNR") == 0)
			fids[fid].track = atoi (value);
		else if (strcmp (label, "CODEC") == 0)
			fids[fid].codec = strdup (value);
		else if (strcmp (label, "TYPE") == 0) {
			if (strcmp (value, "tune") == 0)
				fids[fid].type = TYPE_TUNE;
			else
				fids[fid].type = TYPE_PLAYLIST;
		} else if (strcmp (label, "DURATION") == 0)
			fids[fid].duration = atol (value);
	}
	fclose (file);
}

void strupper (char *string)
{
	unsigned long l = strlen (string), i;
	
	for (i = 0; i < l; i++)
		string[i] = toupper (string[i]);
}

void mirrorPlaylist (char *dir, unsigned long fid)
{
	char *newDir, filename[32], m3uname[1024], *dirTitle;
	FILE *file, *m3u;
	unsigned long subFid;
	
	// Skip unattached items playlist
	if (fid == 17) {
		printf ("Warning: skipping unattached items, if any\n");
		return;
	}	

	newDir = malloc (strlen (dir) + strlen (fids[fid].title) + 2);
	if (newDir == NULL) {
		fprintf (stderr, "Error: could not allocate space for "
			"new directory name\n");
		exit (1);
	}
	strcpy (newDir, dir);
	strcat (newDir, "/");
	if (fids[fid].title == NULL) {
		fprintf (stderr, "Error: playlist without title "
			"encountered\n");
		exit (1);
	}
	dirTitle = strdup (fids[fid].title);
	if (dirTitle == NULL) {
		fprintf (stderr, "Error: could not copy playlist "
			"title\n");
		exit (1);
	}
	sanatize (dirTitle);
	strcat (newDir, dirTitle);
	errno = 0;
	if ((mkdir (newDir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH |
		S_IXOTH) == -1) && (errno != EEXIST)) {
		perror ("Error creating directory");
		exit (1);
	}
	sprintf (m3uname, "%s/%s.m3u", newDir, dirTitle);
	if ((m3u = fopen (m3uname, "w")) == NULL) {
		perror ("Error creating Winamp playlist file");
		exit (1);
	}
	fprintf (m3u, "#EXTM3U\r\n");
	sprintf (filename, "/drive%d/fids/%lx0", fids[fid].disk, fid);
	file = fopen (filename, "r");
	if (file == NULL) {
		perror ("Error opening FID info file");
		exit (1);
	}
	subFid = (getc (file) + 256 * getc (file) +
		256 * 256 * getc (file) + 256 * 256 * 256 * getc (file))
		>> 4;
	while (! feof (file)) {
		if (fids[subFid].type == TYPE_PLAYLIST)
			mirrorPlaylist (newDir, subFid);
		else
			if ((fids[subFid].codec != NULL) &&
				(strlen (fids[subFid].codec) > 0)) {
				char oldname[32], newname[1024] = "",
					title[1024] = "", *artist,
					*songtitle;
				long duration;

				sprintf (oldname, "/drive%d/fids/%lx0",
					fids[subFid].disk, subFid);
				strcpy (newname, newDir);
				strcat (newname, "/");
				if ((fids[subFid].artist != NULL) &&
					(strlen (fids[subFid].artist) >
					0)) {
					strcat (title,
						fids[subFid].artist);
					strcat (title, "-");
				}
				if ((fids[subFid].source != NULL) &&
					(strlen (fids[subFid].source) >
					0)) {
					strcat (title,
						fids[subFid].source);
					strcat (title, "-");
				}
				if (fids[subFid].track != -1) {
					char trackStr[4];
					sprintf (trackStr, "%02d",
						fids[subFid].track);
					strcat (title, trackStr);
					strcat (title, "-");
				}
				if ((fids[subFid].title != NULL) &&
					(strlen (fids[subFid].title) >
					0)) {
					strcat (title,
						fids[subFid].title);
					strcat (title, "-");
				}
				if (strlen (title) == 0)
					sprintf (title, "fid_%lx",
						subFid);
				else
					title[strlen (title) - 1] = '\0';
				strcat (title, ".");
				strcat (title, fids[subFid].codec);
				sanatize (title);
				strcat (newname, title);
				if ((symlink (oldname, newname) == -1) &&
					(errno != EEXIST)) {
					perror ("Error creating symlink");
					printf ("Old name: %s, new "
						"name: %s\n", oldname,
						newname);
					exit (1);
				}
				duration = fids[subFid].duration / 1000;
				artist = fids[subFid].artist;
				songtitle = fids[subFid].title;
				if ((artist == NULL) || (strlen (artist)
					== 0))
					artist = "(Unknown)";
				if ((songtitle == NULL) || (strlen
					(songtitle) == 0))
					songtitle = "(Unknown)";
				errno = 0;
				fprintf (m3u, "#EXTINF:%ld,%s - %s\r\n",
					duration, artist, songtitle);
				fprintf (m3u, "%s\r\n", title);
				if (errno)
				{
					perror ("Error writing Winamp "
						"playlist");
					exit (1);
				}
			}
		subFid = (getc (file) + 256 * getc (file) +
			256 * 256 * getc (file) +
			256 * 256 * 256 * getc (file)) >> 4;
	}
	fclose (file);
	fclose (m3u);
}

void sanatize (unsigned char *string)
{
	unsigned long i, l = strlen (string);
	
	for (i = 0; i < l; i++)
		if ((string[i] < 32) || (string[i] == '?') ||
			(string[i] == '*') || (string[i] == ' ') ||
			(string[i] == '/') || (string[i] == '\\') ||
			(string[i] == '>') || (string[i] == '<') ||
			(string[i] == ':') || (string[i] == '|') ||
			(string[i] == '"'))
			string[i] = '_';
}